@@ -16,7 +16,11 @@ import subprocess, os.path, tempfile
 
 from .. import mlog
 from .. import coredata
-from ..mesonlib import EnvironmentException, version_compare, Popen_safe, listify
+from . import compilers
+from ..mesonlib import (
+    EnvironmentException, version_compare, Popen_safe, listify,
+    for_windows, for_darwin, for_cygwin, for_haiku,
+)
 
 from .compilers import (
     GCC_MINGW,
@@ -88,6 +92,8 @@ class CCompiler(Compiler):
 
     # The default behavior is this, override in MSVC
     def build_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
+        if self.id == 'clang' and self.clang_type == compilers.CLANG_OSX:
+            return self.build_osx_rpath_args(build_dir, rpath_paths, build_rpath)
         return self.build_unix_rpath_args(build_dir, from_dir, rpath_paths, build_rpath, install_rpath)
 
     def get_dependency_gen_args(self, outtarget, outfile):
@@ -277,12 +283,12 @@ class CCompiler(Compiler):
             # Add compile flags needed by dependencies
             args += d.get_compile_args()
             if d.need_threads():
-                args += self.thread_flags()
+                args += self.thread_flags(env)
             if mode == 'link':
                 # Add link flags needed to find dependencies
                 args += d.get_link_args()
                 if d.need_threads():
-                    args += self.thread_link_flags()
+                    args += self.thread_link_flags(env)
         # Select a CRT if needed since we're linking
         if mode == 'link':
             args += self.get_linker_debug_crt_args()
@@ -710,7 +716,47 @@ class CCompiler(Compiler):
                         return False
         raise RuntimeError('BUG: {!r} check failed unexpectedly'.format(n))
 
-    def find_library(self, libname, env, extra_dirs):
+    def get_library_naming(self, env, libtype):
+        '''
+        Get library prefixes and suffixes for the target platform ordered by
+        priority
+        '''
+        stlibext = ['a']
+        # We've always allowed libname to be both `foo` and `libfoo`,
+        # and now people depend on it
+        prefixes = ['lib', '']
+        # Library suffixes and prefixes
+        if for_darwin(env.is_cross_build(), env):
+            shlibext = ['dylib']
+        elif for_windows(env.is_cross_build(), env):
+            # FIXME: .lib files can be import or static so we should read the
+            # file, figure out which one it is, and reject the wrong kind.
+            if self.id == 'msvc':
+                shlibext = ['lib']
+            else:
+                shlibext = ['dll.a', 'lib', 'dll']
+            # Yep, static libraries can also be foo.lib
+            stlibext += ['lib']
+        elif for_cygwin(env.is_cross_build(), env):
+            shlibext = ['dll', 'dll.a']
+            prefixes = ['cyg'] + prefixes
+        else:
+            # Linux/BSDs
+            shlibext = ['so']
+        # Search priority
+        if libtype in ('default', 'shared-static'):
+            suffixes = shlibext + stlibext
+        elif libtype == 'static-shared':
+            suffixes = stlibext + shlibext
+        elif libtype == 'shared':
+            suffixes = shlibext
+        elif libtype == 'static':
+            suffixes = stlibext
+        else:
+            raise AssertionError('BUG: unknown libtype {!r}'.format(libtype))
+        return prefixes, suffixes
+
+    def find_library(self, libname, env, extra_dirs, libtype='default'):
         # These libraries are either built-in or invalid
         if libname in self.ignore_libs:
             return []
@@ -720,27 +766,31 @@ class CCompiler(Compiler):
             extra_dirs = [extra_dirs]
         # Gcc + co seem to prefer builtin lib dirs to -L dirs.
         # Only try to find std libs if no extra dirs specified.
-        if not extra_dirs:
+        if not extra_dirs and libtype == 'default':
             args = ['-l' + libname]
             if self.links(code, env, extra_args=args):
                 return args
-        # Not found? Try to find the library file itself.
+        # Not found or we want to use a specific libtype? Try to find the
+        # library file itself.
         extra_dirs += self.get_library_dirs()
-        suffixes = ['so', 'dylib', 'lib', 'dll', 'a']
+        prefixes, suffixes = self.get_library_naming(env, libtype)
+        # Triply-nested loop!
         for d in extra_dirs:
             for suffix in suffixes:
-                trial = os.path.join(d, 'lib' + libname + '.' + suffix)
-                if os.path.isfile(trial):
-                    return [trial]
-                trial2 = os.path.join(d, libname + '.' + suffix)
-                if os.path.isfile(trial2):
-                    return [trial2]
+                for prefix in prefixes:
+                    trial = os.path.join(d, prefix + libname + '.' + suffix)
+                    if os.path.isfile(trial):
+                        return [trial]
         return None
 
-    def thread_flags(self):
+    def thread_flags(self, env):
+        if for_haiku(self.is_cross, env):
+            return []
         return ['-pthread']
 
-    def thread_link_flags(self):
+    def thread_link_flags(self, env):
+        if for_haiku(self.is_cross, env):
+            return []
         return ['-pthread']
 
     def has_multi_arguments(self, args, env):
@@ -772,6 +822,12 @@ class ClangCCompiler(ClangCompiler, CCompiler):
     def get_option_link_args(self, options):
         return []
 
+    def get_linker_always_args(self):
+        basic = super().get_linker_always_args()
+        if self.clang_type == compilers.CLANG_OSX:
+            return basic + ['-Wl,-headerpad_max_install_names']
+        return basic
+
 
 class GnuCCompiler(GnuCompiler, CCompiler):
     def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None):
@@ -789,8 +845,8 @@ class GnuCCompiler(GnuCompiler, CCompiler):
                                                   'none')}
         if self.gcc_type == GCC_MINGW:
             opts.update({
-                'c_winlibs': coredata.UserStringArrayOption('c_winlibs', 'Standard Win libraries to link against',
-                                                            gnu_winlibs), })
+                'c_winlibs': coredata.UserArrayOption('c_winlibs', 'Standard Win libraries to link against',
+                                                      gnu_winlibs), })
         return opts
 
     def get_option_compile_args(self, options):
@@ -955,16 +1011,16 @@ class VisualStudioCCompiler(CCompiler):
         return []
 
     # FIXME, no idea what these should be.
-    def thread_flags(self):
+    def thread_flags(self, env):
         return []
 
-    def thread_link_flags(self):
+    def thread_link_flags(self, env):
         return []
 
     def get_options(self):
-        return {'c_winlibs': coredata.UserStringArrayOption('c_winlibs',
-                                                            'Windows libs to link against.',
-                                                            msvc_winlibs)
+        return {'c_winlibs': coredata.UserArrayOption('c_winlibs',
+                                                      'Windows libs to link against.',
+                                                      msvc_winlibs)
                 }
 
     def get_option_link_args(self, options):
