@@ -273,7 +273,7 @@ def sanitizer_compile_args(value):
     if value == 'none':
         return []
     args = ['-fsanitize=' + value]
-    if value == 'address':
+    if 'address' in value: # For -fsanitize=address,undefined
         args.append('-fno-omit-frame-pointer')
     return args
 
@@ -451,9 +451,9 @@ class CompilerArgs(list):
         Returns whether the argument can be safely de-duped. This is dependent
         on three things:
 
-        a) Whether an argument can be 'overriden' by a later argument.  For
+        a) Whether an argument can be 'overridden' by a later argument.  For
            example, -DFOO defines FOO and -UFOO undefines FOO. In this case, we
-           can safely remove the previous occurance and add a new one. The same
+           can safely remove the previous occurrence and add a new one. The same
            is true for include paths and library paths with -I and -L. For
            these we return `2`. See `dedup2_prefixes` and `dedup2_args`.
         b) Arguments that once specified cannot be undone, such as `-c` or
@@ -511,10 +511,10 @@ class CompilerArgs(list):
                     continue
                 i = self.index(each)
                 if group_start < 0:
-                    # First occurance of a library
+                    # First occurrence of a library
                     group_start = i
             if group_start >= 0:
-                # Last occurance of a library
+                # Last occurrence of a library
                 self.insert(i + 1, '-Wl,--end-group')
                 self.insert(group_start, '-Wl,--start-group')
         return self.compiler.unix_args_to_native(self)
@@ -548,15 +548,15 @@ class CompilerArgs(list):
             raise TypeError('can only concatenate list (not "{}") to list'.format(args))
         for arg in args:
             # If the argument can be de-duped, do it either by removing the
-            # previous occurance of it and adding a new one, or not adding the
-            # new occurance.
+            # previous occurrence of it and adding a new one, or not adding the
+            # new occurrence.
             dedup = self._can_dedup(arg)
             if dedup == 1:
                 # Argument already exists and adding a new instance is useless
                 if arg in self or arg in pre or arg in post:
                     continue
             if dedup == 2:
-                # Remove all previous occurances of the arg and add it anew
+                # Remove all previous occurrences of the arg and add it anew
                 if arg in self:
                     self.remove(arg)
                 if arg in pre:
@@ -605,7 +605,7 @@ class Compiler:
             self.exelist = exelist
         else:
             raise TypeError('Unknown argument to Compiler')
-        # In case it's been overriden by a child class already
+        # In case it's been overridden by a child class already
         if not hasattr(self, 'file_suffixes'):
             self.file_suffixes = lang_suffixes[self.language]
         if not hasattr(self, 'can_compile_suffixes'):
@@ -817,6 +817,16 @@ class Compiler:
     def get_instruction_set_args(self, instruction_set):
         return None
 
+    def build_osx_rpath_args(self, build_dir, rpath_paths, build_rpath):
+        if not rpath_paths and not build_rpath:
+            return []
+        # On OSX, rpaths must be absolute.
+        abs_rpaths = [os.path.join(build_dir, p) for p in rpath_paths]
+        if build_rpath != '':
+            abs_rpaths.append(build_rpath)
+        args = ['-Wl,-rpath,' + rp for rp in abs_rpaths]
+        return args
+
     def build_unix_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
         if not rpath_paths and not install_rpath and not build_rpath:
             return []
@@ -833,7 +843,9 @@ class Compiler:
         paths = ':'.join([os.path.join('$ORIGIN', p) for p in rel_rpaths])
         # Build_rpath is used as-is (it is usually absolute).
         if build_rpath != '':
-            paths += ':' + build_rpath
+            if paths != '':
+                paths += ':'
+            paths += build_rpath
         if len(paths) < len(install_rpath):
             padding = 'X' * (len(install_rpath) - len(paths))
             if not paths:
@@ -877,7 +889,11 @@ def get_gcc_soname_args(gcc_type, prefix, shlib_name, suffix, path, soversion, i
     elif gcc_type == GCC_OSX:
         if is_shared_module:
             return []
-        return ['-install_name', os.path.join(path, 'lib' + shlib_name + '.dylib')]
+        install_name = prefix + shlib_name
+        if soversion is not None:
+            install_name += '.' + soversion
+        install_name += '.dylib'
+        return ['-install_name', os.path.join('@rpath', install_name)]
     else:
         raise RuntimeError('Not implemented yet.')
 
@@ -923,11 +939,15 @@ def get_largefile_args(compiler):
 def gnulike_default_include_dirs(compiler, lang):
     if lang == 'cpp':
         lang = 'c++'
+    env = os.environ.copy()
+    env["LC_ALL"] = 'C'
+    cmd = compiler + ['-x{}'.format(lang), '-E', '-v', '-']
     p = subprocess.Popen(
-        compiler + ['-x{}'.format(lang), '-E', '-v', '-'],
+        cmd,
         stdin=subprocess.DEVNULL,
         stderr=subprocess.PIPE,
-        stdout=subprocess.PIPE
+        stdout=subprocess.PIPE,
+        env=env
     )
     stderr = p.stderr.read().decode('utf-8')
     parse_state = 0
@@ -946,6 +966,8 @@ def gnulike_default_include_dirs(compiler, lang):
                 break
             else:
                 paths.append(line[1:])
+    if len(paths) == 0:
+        mlog.warning('No include directory found parsing "{cmd}" output'.format(cmd=" ".join(cmd)))
     return paths
 
 class GnuCompiler:
