@@ -16,14 +16,19 @@
 # Custom logic for several other packages are in separate files.
 
 import os
-import shutil
-import stat
 import sys
+import stat
+import shlex
+import shutil
+import textwrap
 from enum import Enum
+from pathlib import PurePath
 
 from .. import mlog
 from .. import mesonlib
-from ..mesonlib import MesonException, Popen_safe, flatten, version_compare_many
+from ..mesonlib import (
+    MesonException, Popen_safe, version_compare_many, version_compare, listify
+)
 
 
 # These must be defined in this file to avoid cyclical references.
@@ -42,14 +47,17 @@ class DependencyMethods(Enum):
     QMAKE = 'qmake'
     # Just specify the standard link arguments, assuming the operating system provides the library.
     SYSTEM = 'system'
-    # Detect using sdl2-config
-    SDLCONFIG = 'sdlconfig'
-    # Detect using pcap-config
-    PCAPCONFIG = 'pcap-config'
     # This is only supported on OSX - search the frameworks directory by name.
     EXTRAFRAMEWORK = 'extraframework'
     # Detect using the sysconfig module.
     SYSCONFIG = 'sysconfig'
+    # Specify using a "program"-config style tool
+    CONFIG_TOOL = 'config-tool'
+    # For backewards compatibility
+    SDLCONFIG = 'sdlconfig'
+    CUPSCONFIG = 'cups-config'
+    PCAPCONFIG = 'pcap-config'
+    LIBWMFCONFIG = 'libwmf-config'
 
 
 class Dependency:
@@ -67,6 +75,16 @@ class Dependency:
             raise DependencyException('method {!r} is invalid'.format(method))
         method = DependencyMethods(method)
 
+        # This sets per-too config methods which are deprecated to to the new
+        # generic CONFIG_TOOL value.
+        if method in [DependencyMethods.SDLCONFIG, DependencyMethods.CUPSCONFIG,
+                      DependencyMethods.PCAPCONFIG, DependencyMethods.LIBWMFCONFIG]:
+            mlog.warning(textwrap.dedent("""\
+                Configuration method {} has been deprecated in favor of
+                'config-tool'. This will be removed in a future version of
+                meson.""".format(method)))
+            method = DependencyMethods.CONFIG_TOOL
+
         # Set the detection method. If the method is set to auto, use any available method.
         # If method is set to a specific string, allow only that detection method.
         if method == DependencyMethods.AUTO:
@@ -77,7 +95,7 @@ class Dependency:
             raise DependencyException(
                 'Unsupported detection method: {}, allowed methods are {}'.format(
                     method.value,
-                    mlog.format_list(map(lambda x: x.value, [DependencyMethods.AUTO] + self.get_methods()))))
+                    mlog.format_list([x.value for x in [DependencyMethods.AUTO] + self.get_methods()])))
 
     def __repr__(self):
         s = '<{0} {1}: {2}>'
@@ -112,8 +130,11 @@ class Dependency:
     def need_threads(self):
         return False
 
-    def get_pkgconfig_variable(self, variable_name):
-        raise NotImplementedError('{!r} is not a pkgconfig dependency'.format(self.name))
+    def get_pkgconfig_variable(self, variable_name, kwargs):
+        raise DependencyException('{!r} is not a pkgconfig dependency'.format(self.name))
+
+    def get_configtool_variable(self, variable_name):
+        raise DependencyException('{!r} is not a config-tool dependency'.format(self.name))
 
 
 class InternalDependency(Dependency):
@@ -128,6 +149,14 @@ class InternalDependency(Dependency):
         self.sources = sources
         self.ext_deps = ext_deps
 
+    def get_pkgconfig_variable(self, variable_name, kwargs):
+        raise DependencyException('Method "get_pkgconfig_variable()" is '
+                                  'invalid for an internal dependency')
+
+    def get_configtool_variable(self, variable_name):
+        raise DependencyException('Method "get_configtool_variable()" is '
+                                  'invalid for an internal dependency')
+
 
 class ExternalDependency(Dependency):
     def __init__(self, type_name, environment, language, kwargs):
@@ -136,9 +165,6 @@ class ExternalDependency(Dependency):
         self.name = type_name # default
         self.is_found = False
         self.language = language
-        if language and language not in self.env.coredata.compilers:
-            m = self.name.capitalize() + ' requires a {} compiler'
-            raise DependencyException(m.format(language.capitalize()))
         self.version_reqs = kwargs.get('version', None)
         self.required = kwargs.get('required', True)
         self.silent = kwargs.get('silent', False)
@@ -156,19 +182,153 @@ class ExternalDependency(Dependency):
             compilers = self.env.coredata.cross_compilers
         else:
             compilers = self.env.coredata.compilers
-        self.compiler = compilers.get(self.language or 'c', None)
+        # Set the compiler for this dependency if a language is specified,
+        # else try to pick something that looks usable.
+        if self.language:
+            if self.language not in compilers:
+                m = self.name.capitalize() + ' requires a {} compiler'
+                raise DependencyException(m.format(self.language.capitalize()))
+            self.compiler = compilers[self.language]
+        else:
+            # Try to find a compiler that this dependency can use for compiler
+            # checks. It's ok if we don't find one.
+            for lang in ('c', 'cpp', 'objc', 'objcpp', 'fortran', 'd'):
+                self.compiler = compilers.get(lang, None)
+                if self.compiler:
+                    break
 
     def get_compiler(self):
         return self.compiler
 
 
+class ConfigToolDependency(ExternalDependency):
+
+    """Class representing dependencies found using a config tool."""
+
+    tools = None
+    tool_name = None
+
+    def __init__(self, name, environment, language, kwargs):
+        super().__init__('config-tool', environment, language, kwargs)
+        self.name = name
+        self.tools = listify(kwargs.get('tools', self.tools))
+
+        req_version = kwargs.get('version', None)
+        tool, version = self.find_config(req_version)
+        self.config = tool
+        self.is_found = self.report_config(version, req_version)
+        if not self.is_found:
+            self.config = None
+            return
+        self.version = version
+
+    @classmethod
+    def factory(cls, name, environment, language, kwargs, tools, tool_name):
+        """Constructor for use in dependencies that can be found multiple ways.
+
+        In addition to the standard constructor values, this constructor sets
+        the tool_name and tools values of the instance.
+        """
+        # This deserves some explanation, because metaprogramming is hard.
+        # This uses type() to create a dynamic subclass of ConfigToolDependency
+        # with the tools and tool_name class attributes set, this class is then
+        # instantiated and returned. The reduce function (method) is also
+        # attached, since python's pickle module won't be able to do anything
+        # with this dynamically generated class otherwise.
+        def reduce(_):
+            return (cls.factory,
+                    (name, environment, language, kwargs, tools, tool_name))
+        sub = type('{}Dependency'.format(name.capitalize()), (cls, ),
+                   {'tools': tools, 'tool_name': tool_name, '__reduce__': reduce})
+
+        return sub(name, environment, language, kwargs)
+
+    def find_config(self, versions=None):
+        """Helper method that searchs for config tool binaries in PATH and
+        returns the one that best matches the given version requirements.
+        """
+        if not isinstance(versions, list) and versions is not None:
+            versions = listify(versions)
+
+        best_match = (None, None)
+        for tool in self.tools:
+            try:
+                p, out = Popen_safe([tool, '--version'])[:2]
+            except (FileNotFoundError, PermissionError):
+                continue
+            if p.returncode != 0:
+                continue
+
+            out = out.strip()
+            # Some tools, like pcap-config don't supply a version, but also
+            # dont fail with --version, in that case just assume that there is
+            # only one verison and return it.
+            if not out:
+                return (tool, 'none')
+            if versions:
+                is_found = version_compare_many(out, versions)[0]
+                # This allows returning a found version without a config tool,
+                # which is useful to inform the user that you found version x,
+                # but y was required.
+                if not is_found:
+                    tool = None
+            if best_match[1]:
+                if version_compare(out, '> {}'.format(best_match[1])):
+                    best_match = (tool, out)
+            else:
+                best_match = (tool, out)
+
+        return best_match
+
+    def report_config(self, version, req_version):
+        """Helper method to print messages about the tool."""
+        if self.config is None:
+            if version is not None:
+                mlog.log('found {} {!r} but need:'.format(self.tool_name, version),
+                         req_version)
+            else:
+                mlog.log("No {} found; can't detect dependency".format(self.tool_name))
+            mlog.log('Dependency {} found:'.format(self.name), mlog.red('NO'))
+            if self.required:
+                raise DependencyException('Dependency {} not found'.format(self.name))
+            return False
+        mlog.log('Found {}:'.format(self.tool_name), mlog.bold(shutil.which(self.config)),
+                 '({})'.format(version))
+        mlog.log('Dependency {} found:'.format(self.name), mlog.green('YES'))
+        return True
+
+    def get_config_value(self, args, stage):
+        p, out, err = Popen_safe([self.config] + args)
+        if p.returncode != 0:
+            if self.required:
+                raise DependencyException(
+                    'Could not generate {} for {}.\n{}'.format(
+                        stage, self.name, err))
+            return []
+        return shlex.split(out)
+
+    def get_methods(self):
+        return [DependencyMethods.AUTO, DependencyMethods.CONFIG_TOOL]
+
+    def get_configtool_variable(self, variable_name):
+        p, out, _ = Popen_safe([self.config, '--{}'.format(variable_name)])
+        if p.returncode != 0:
+            if self.required:
+                raise DependencyException(
+                    'Could not get variable "{}" for dependency {}'.format(
+                        variable_name, self.name))
+        variable = out.strip()
+        mlog.debug('Got config-tool variable {} : {}'.format(variable_name, variable))
+        return variable
+
+
 class PkgConfigDependency(ExternalDependency):
     # The class's copy of the pkg-config path. Avoids having to search for it
     # multiple times in the same Meson invocation.
     class_pkgbin = None
 
-    def __init__(self, name, environment, kwargs):
-        super().__init__('pkgconfig', environment, None, kwargs)
+    def __init__(self, name, environment, kwargs, language=None):
+        super().__init__('pkgconfig', environment, language, kwargs)
         self.name = name
         self.is_libtool = False
         # Store a copy of the pkg-config path on the object itself so it is
@@ -254,28 +414,78 @@ class PkgConfigDependency(ExternalDependency):
         return s.format(self.__class__.__name__, self.name, self.is_found,
                         self.version_reqs)
 
-    def _call_pkgbin(self, args):
-        p, out = Popen_safe([self.pkgbin] + args, env=os.environ)[0:2]
+    def _call_pkgbin(self, args, env=None):
+        if not env:
+            env = os.environ
+        p, out = Popen_safe([self.pkgbin] + args, env=env)[0:2]
         return p.returncode, out.strip()
 
+    def _convert_mingw_paths(self, args):
+        '''
+        Both MSVC and native Python on Windows cannot handle MinGW-esque /c/foo
+        paths so convert them to C:/foo. We cannot resolve other paths starting
+        with / like /home/foo so leave them as-is so that the user gets an
+        error/warning from the compiler/linker.
+        '''
+        if not mesonlib.is_windows():
+            return args
+        converted = []
+        for arg in args:
+            pargs = []
+            # Library search path
+            if arg.startswith('-L/'):
+                pargs = PurePath(arg[2:]).parts
+                tmpl = '-L{}:/{}'
+            elif arg.startswith('-I/'):
+                pargs = PurePath(arg[2:]).parts
+                tmpl = '-I{}:/{}'
+            # Full path to library or .la file
+            elif arg.startswith('/'):
+                pargs = PurePath(arg).parts
+                tmpl = '{}:/{}'
+            if len(pargs) > 1 and len(pargs[1]) == 1:
+                arg = tmpl.format(pargs[1], '/'.join(pargs[2:]))
+            converted.append(arg)
+        return converted
+
     def _set_cargs(self):
         ret, out = self._call_pkgbin(['--cflags', self.name])
         if ret != 0:
             raise DependencyException('Could not generate cargs for %s:\n\n%s' %
                                       (self.name, out))
-        self.compile_args = out.split()
+        self.compile_args = self._convert_mingw_paths(shlex.split(out))
 
     def _set_libs(self):
+        env = None
         libcmd = [self.name, '--libs']
         if self.static:
             libcmd.append('--static')
-        ret, out = self._call_pkgbin(libcmd)
+            # Force pkg-config to output -L fields even if they are system
+            # paths so we can do manual searching with cc.find_library() later.
+            env = os.environ.copy()
+            env['PKG_CONFIG_ALLOW_SYSTEM_LIBS'] = '1'
+        ret, out = self._call_pkgbin(libcmd, env=env)
         if ret != 0:
             raise DependencyException('Could not generate libs for %s:\n\n%s' %
                                       (self.name, out))
         self.link_args = []
-        for lib in out.split():
-            if lib.endswith(".la"):
+        libpaths = []
+        for lib in self._convert_mingw_paths(shlex.split(out)):
+            # If we want to use only static libraries, we have to look for the
+            # file ourselves instead of depending on the compiler to find it
+            # with -lfoo or foo.lib. However, we can only do this if we already
+            # have some library paths gathered.
+            if self.static:
+                if lib.startswith('-L'):
+                    libpaths.append(lib[2:])
+                    continue
+                elif lib.startswith('-l') and libpaths:
+                    args = self.compiler.find_library(lib[2:], self.env, libpaths, libtype='static')
+                    if not args or len(args) < 1:
+                        raise DependencyException('Static library not found for {!r}'
+                                                  ''.format(lib[2:]))
+                    lib = args[0]
+            elif lib.endswith(".la"):
                 shared_libname = self.extract_libtool_shlib(lib)
                 shared_lib = os.path.join(os.path.dirname(lib), shared_libname)
                 if not os.path.exists(shared_lib):
@@ -289,8 +499,20 @@ class PkgConfigDependency(ExternalDependency):
                 self.is_libtool = True
             self.link_args.append(lib)
 
-    def get_pkgconfig_variable(self, variable_name):
-        ret, out = self._call_pkgbin(['--variable=' + variable_name, self.name])
+    def get_pkgconfig_variable(self, variable_name, kwargs):
+        options = ['--variable=' + variable_name, self.name]
+
+        if 'define_variable' in kwargs:
+            definition = kwargs.get('define_variable', [])
+            if not isinstance(definition, list):
+                raise MesonException('define_variable takes a list')
+
+            if len(definition) != 2 or not all(isinstance(i, str) for i in definition):
+                raise MesonException('define_variable must be made up of 2 strings for VARIABLENAME and VARIABLEVALUE')
+
+            options = ['--define-variable=' + '='.join(definition)] + options
+
+        ret, out = self._call_pkgbin(options)
         variable = ''
         if ret != 0:
             if self.required:
@@ -367,15 +589,12 @@ class PkgConfigDependency(ExternalDependency):
 
 
 class ExternalProgram:
-    windows_exts = ('exe', 'msc', 'com', 'bat')
+    windows_exts = ('exe', 'msc', 'com', 'bat', 'cmd')
 
     def __init__(self, name, command=None, silent=False, search_dir=None):
         self.name = name
         if command is not None:
-            if not isinstance(command, list):
-                self.command = [command]
-            else:
-                self.command = command
+            self.command = listify(command)
         else:
             self.command = self._search(name, search_dir)
         if not silent:
@@ -400,7 +619,11 @@ class ExternalProgram:
             with open(script) as f:
                 first_line = f.readline().strip()
             if first_line.startswith('#!'):
-                commands = first_line[2:].split('#')[0].strip().split()
+                # In a shebang, everything before the first space is assumed to
+                # be the command to run and everything after the first space is
+                # the single argument to pass to that command. So we must split
+                # exactly once.
+                commands = first_line[2:].split('#')[0].strip().split(maxsplit=1)
                 if mesonlib.is_windows():
                     # Windows does not have UNIX paths so remove them,
                     # but don't remove Windows paths
@@ -410,10 +633,21 @@ class ExternalProgram:
                         commands = commands[1:]
                     # Windows does not ship python3.exe, but we know the path to it
                     if len(commands) > 0 and commands[0] == 'python3':
-                        commands[0] = sys.executable
+                        commands = mesonlib.python_command + commands[1:]
+                elif mesonlib.is_haiku():
+                    # Haiku does not have /usr, but a lot of scripts assume that
+                    # /usr/bin/env always exists. Detect that case and run the
+                    # script with the interpreter after it.
+                    if commands[0] == '/usr/bin/env':
+                        commands = commands[1:]
+                    # We know what python3 is, we're running on it
+                    if len(commands) > 0 and commands[0] == 'python3':
+                        commands = mesonlib.python_command + commands[1:]
                 return commands + [script]
-        except Exception:
+        except Exception as e:
+            mlog.debug(e)
             pass
+        mlog.debug('Unusable script {!r}'.format(script))
         return False
 
     def _is_executable(self, path):
@@ -444,21 +678,17 @@ class ExternalProgram:
                         return [trial_ext]
         return False
 
-    def _search(self, name, search_dir):
+    def _search_windows_special_cases(self, name, command):
         '''
-        Search in the specified dir for the specified executable by name
-        and if not found search in PATH
+        Lots of weird Windows quirks:
+        1. PATH search for @name returns files with extensions from PATHEXT,
+           but only self.windows_exts are executable without an interpreter.
+        2. @name might be an absolute path to an executable, but without the
+           extension. This works inside MinGW so people use it a lot.
+        3. The script is specified without an extension, in which case we have
+           to manually search in PATH.
+        4. More special-casing for the shebang inside the script.
         '''
-        commands = self._search_dir(name, search_dir)
-        if commands:
-            return commands
-        # Do a standard search in PATH
-        command = shutil.which(name)
-        if not mesonlib.is_windows():
-            # On UNIX-like platforms, shutil.which() is enough to find
-            # all executables whether in PATH or with an absolute path
-            return [command]
-        # HERE BEGINS THE TERROR OF WINDOWS
         if command:
             # On Windows, even if the PATH search returned a full path, we can't be
             # sure that it can be run directly if it's not a native executable.
@@ -472,25 +702,41 @@ class ExternalProgram:
             commands = self._shebang_to_cmd(command)
             if commands:
                 return commands
-        else:
-            # Maybe the name is an absolute path to a native Windows
-            # executable, but without the extension. This is technically wrong,
-            # but many people do it because it works in the MinGW shell.
-            if os.path.isabs(name):
-                for ext in self.windows_exts:
-                    command = '{}.{}'.format(name, ext)
-                    if os.path.exists(command):
-                        return [command]
-            # On Windows, interpreted scripts must have an extension otherwise they
-            # cannot be found by a standard PATH search. So we do a custom search
-            # where we manually search for a script with a shebang in PATH.
-            search_dirs = os.environ.get('PATH', '').split(';')
-            for search_dir in search_dirs:
-                commands = self._search_dir(name, search_dir)
-                if commands:
-                    return commands
+            return [None]
+        # Maybe the name is an absolute path to a native Windows
+        # executable, but without the extension. This is technically wrong,
+        # but many people do it because it works in the MinGW shell.
+        if os.path.isabs(name):
+            for ext in self.windows_exts:
+                command = '{}.{}'.format(name, ext)
+                if os.path.exists(command):
+                    return [command]
+        # On Windows, interpreted scripts must have an extension otherwise they
+        # cannot be found by a standard PATH search. So we do a custom search
+        # where we manually search for a script with a shebang in PATH.
+        search_dirs = os.environ.get('PATH', '').split(';')
+        for search_dir in search_dirs:
+            commands = self._search_dir(name, search_dir)
+            if commands:
+                return commands
         return [None]
 
+    def _search(self, name, search_dir):
+        '''
+        Search in the specified dir for the specified executable by name
+        and if not found search in PATH
+        '''
+        commands = self._search_dir(name, search_dir)
+        if commands:
+            return commands
+        # Do a standard search in PATH
+        command = shutil.which(name)
+        if mesonlib.is_windows():
+            return self._search_windows_special_cases(name, command)
+        # On UNIX-like platforms, shutil.which() is enough to find
+        # all executables whether in PATH or with an absolute path
+        return [command]
+
     def found(self):
         return self.command[0] is not None
 
@@ -587,7 +833,7 @@ class ExtraFrameworkDependency(ExternalDependency):
 
 def get_dep_identifier(name, kwargs, want_cross):
     # Need immutable objects since the identifier will be used as a dict key
-    version_reqs = flatten(kwargs.get('version', []))
+    version_reqs = listify(kwargs.get('version', []))
     if isinstance(version_reqs, list):
         version_reqs = frozenset(version_reqs)
     identifier = (name, version_reqs, want_cross)
@@ -600,7 +846,7 @@ def get_dep_identifier(name, kwargs, want_cross):
             continue
         # All keyword arguments are strings, ints, or lists (or lists of lists)
         if isinstance(value, list):
-            value = frozenset(flatten(value))
+            value = frozenset(listify(value))
         identifier += (key, value)
     return identifier
 
@@ -641,3 +887,14 @@ def find_external_dependency(name, env, kwargs):
         raise pkg_exc
     mlog.log('Dependency', mlog.bold(name), 'found:', mlog.red('NO'))
     return pkgdep
+
+
+def strip_system_libdirs(environment, link_args):
+    """Remove -L<system path> arguments.
+
+    leaving these in will break builds where a user has a version of a library
+    in the system path, and a different version not in the system path if they
+    want to link against the non-system path version.
+    """
+    exclude = {'-L{}'.format(p) for p in environment.get_compiler_system_dirs()}
+    return [l for l in link_args if l not in exclude]
