@@ -16,13 +16,17 @@
 # development purposes, such as testing, debugging, etc..
 
 import os
-import shlex
+import re
 import shutil
 
 from .. import mlog
 from .. import mesonlib
 from ..mesonlib import version_compare, Popen_safe, stringlistify, extract_as_list
-from .base import DependencyException, ExternalDependency, PkgConfigDependency
+from .base import (
+    DependencyException, ExternalDependency, PkgConfigDependency,
+    strip_system_libdirs, ConfigToolDependency,
+)
+
 
 class GTestDependency(ExternalDependency):
     def __init__(self, environment, kwargs):
@@ -107,122 +111,118 @@ class GMockDependency(ExternalDependency):
         self.is_found = False
 
 
-class LLVMDependency(ExternalDependency):
+class LLVMDependency(ConfigToolDependency):
     """
     LLVM uses a special tool, llvm-config, which has arguments for getting
     c args, cxx args, and ldargs as well as version.
     """
 
     # Ordered list of llvm-config binaries to try. Start with base, then try
-    # newest back to oldest (3.5 is abitrary), and finally the devel version.
-    # Please note that llvm-config-5.0 is a development snapshot and it should
+    # newest back to oldest (3.5 is arbitrary), and finally the devel version.
+    # Please note that llvm-config-6.0 is a development snapshot and it should
     # not be moved to the beginning of the list. The only difference between
-    # llvm-config-5.0 and llvm-config-devel is that the former is used by
+    # llvm-config-6.0 and llvm-config-devel is that the former is used by
     # Debian and the latter is used by FreeBSD.
-    llvm_config_bins = [
+    tools = [
         'llvm-config', # base
-        'llvm-config-4.0', 'llvm-config40', # latest stable release
-        'llvm-config-3.9', 'llvm-config39', # old stable releases
+        'llvm-config-5.0', 'llvm-config50', # latest stable release
+        'llvm-config-4.0', 'llvm-config40', # old stable releases
+        'llvm-config-3.9', 'llvm-config39',
         'llvm-config-3.8', 'llvm-config38',
         'llvm-config-3.7', 'llvm-config37',
         'llvm-config-3.6', 'llvm-config36',
         'llvm-config-3.5', 'llvm-config35',
-        'llvm-config-5.0', 'llvm-config-devel', # development snapshot
+        'llvm-config-6.0', 'llvm-config-devel', # development snapshot
     ]
+    tool_name = 'llvm-config'
     __cpp_blacklist = {'-DNDEBUG'}
 
     def __init__(self, environment, kwargs):
         # It's necessary for LLVM <= 3.8 to use the C++ linker. For 3.9 and 4.0
         # the C linker works fine if only using the C API.
-        super().__init__('llvm-config', environment, 'cpp', kwargs)
-        self.modules = []
-        self.llvmconfig = None
-        self.__best_found = None
-        # FIXME: Support multiple version requirements ala PkgConfigDependency
-        req_version = kwargs.get('version', None)
-        self.check_llvmconfig(req_version)
-        if self.llvmconfig is None:
-            if self.__best_found is not None:
-                mlog.log('found {!r} but need:'.format(self.__best_found),
-                         req_version)
-            else:
-                mlog.log("No llvm-config found; can't detect dependency")
-            mlog.log('Dependency LLVM found:', mlog.red('NO'))
-            if self.required:
-                raise DependencyException('Dependency LLVM not found')
+        super().__init__('config-tool', environment, 'cpp', kwargs)
+        self.provided_modules = []
+        self.required_modules = set()
+        if not self.is_found:
             return
+        self.static = kwargs.get('static', False)
 
-        p, out, err = Popen_safe([self.llvmconfig, '--version'])
-        if p.returncode != 0:
-            mlog.debug('stdout: {}\nstderr: {}'.format(out, err))
-            if self.required:
-                raise DependencyException('Dependency LLVM not found')
-            mlog.log('Dependency LLVM found:', mlog.red('NO'))
-            return
-
-        mlog.log('Dependency LLVM found:', mlog.green('YES'))
-        self.is_found = True
-
-        # Currently meson doesn't really atempt to handle pre-release versions,
+        # Currently meson doesn't really attempt to handle pre-release versions,
         # so strip the 'svn' off the end, since it will probably cuase problems
         # for users who want the patch version.
-        self.version = out.strip().rstrip('svn')
+        self.version = self.version.rstrip('svn')
 
-        p, out = Popen_safe(
-            [self.llvmconfig, '--libs', '--ldflags'])[:2]
-        if p.returncode != 0:
-            raise DependencyException('Could not generate libs for LLVM.')
-        self.link_args = shlex.split(out)
+        self.provided_modules = self.get_config_value(['--components'], 'modules')
+        modules = stringlistify(extract_as_list(kwargs, 'modules'))
+        self.check_components(modules)
+        opt_modules = stringlistify(extract_as_list(kwargs, 'optional_modules'))
+        self.check_components(opt_modules, required=False)
 
-        p, out = Popen_safe([self.llvmconfig, '--cppflags'])[:2]
-        if p.returncode != 0:
-            raise DependencyException('Could not generate includedir for LLVM.')
-        cargs = mesonlib.OrderedSet(shlex.split(out))
+        cargs = set(self.get_config_value(['--cppflags'], 'compile_args'))
         self.compile_args = list(cargs.difference(self.__cpp_blacklist))
 
-        p, out = Popen_safe([self.llvmconfig, '--components'])[:2]
-        if p.returncode != 0:
-            raise DependencyException('Could not generate modules for LLVM.')
-        self.modules = shlex.split(out)
+        if version_compare(self.version, '>= 3.9'):
+            self._set_new_link_args()
+        else:
+            self._set_old_link_args()
+        self.link_args = strip_system_libdirs(environment, self.link_args)
 
-        modules = stringlistify(extract_as_list(kwargs, 'modules'))
-        for mod in sorted(set(modules)):
-            if mod not in self.modules:
-                mlog.log('LLVM module', mod, 'found:', mlog.red('NO'))
-                self.is_found = False
-                if self.required:
-                    raise DependencyException(
-                        'Could not find required LLVM Component: {}'.format(mod))
-            else:
-                mlog.log('LLVM module', mod, 'found:', mlog.green('YES'))
+    def _set_new_link_args(self):
+        """How to set linker args for LLVM versions >= 3.9"""
+        link_args = ['--link-static', '--system-libs'] if self.static else ['--link-shared']
+        self.link_args = self.get_config_value(
+            ['--libs', '--ldflags'] + link_args + list(self.required_modules),
+            'link_args')
 
-    def check_llvmconfig(self, version_req):
-        """Try to find the highest version of llvm-config."""
-        for llvmconfig in self.llvm_config_bins:
-            try:
-                p, out = Popen_safe([llvmconfig, '--version'])[0:2]
-                out = out.strip()
-                if p.returncode != 0:
-                    continue
-                if version_req:
-                    if version_compare(out, version_req, strict=True):
-                        if self.__best_found and version_compare(
-                                out, '<={}'.format(self.__best_found), strict=True):
-                            continue
-                        self.__best_found = out
-                        self.llvmconfig = llvmconfig
-                else:
-                    # If no specific version is requested use the first version
-                    # found, since that should be the best.
-                    self.__best_found = out
-                    self.llvmconfig = llvmconfig
+    def _set_old_link_args(self):
+        """Setting linker args for older versions of llvm.
+
+        Old versions of LLVM bring an extra level of insanity with them.
+        llvm-config will provide the correct arguments for static linking, but
+        not for shared-linnking, we have to figure those out ourselves, because
+        of course we do.
+        """
+        if self.static:
+            self.link_args = self.get_config_value(
+                ['--libs', '--ldflags', '--system-libs'] + list(self.required_modules),
+                'link_args')
+        else:
+            # llvm-config will provide arguments for static linking, so we get
+            # to figure out for ourselves what to link with. We'll do that by
+            # checking in the directory provided by --libdir for a library
+            # called libLLVM-<ver>.(so|dylib|dll)
+            libdir = self.get_config_value(['--libdir'], 'link_args')[0]
+
+            expected_name = 'libLLVM-{}'.format(self.version)
+            re_name = re.compile(r'{}.(so|dll|dylib)'.format(expected_name))
+
+            for file_ in os.listdir(libdir):
+                if re_name.match(file_):
+                    self.link_args = ['-L{}'.format(libdir),
+                                      '-l{}'.format(os.path.splitext(file_.lstrip('lib'))[0])]
                     break
-            except (FileNotFoundError, PermissionError):
-                pass
-        if self.__best_found:
-            mlog.log('Found llvm-config:',
-                     mlog.bold(shutil.which(self.llvmconfig)),
-                     '({})'.format(out.strip()))
+            else:
+                raise DependencyException(
+                    'Could not find a dynamically linkable library for LLVM.')
+
+    def check_components(self, modules, required=True):
+        """Check for llvm components (modules in meson terms).
+
+        The required option is whether the module is required, not whether LLVM
+        is required.
+        """
+        for mod in sorted(set(modules)):
+            if mod not in self.provided_modules:
+                mlog.log('LLVM module', mod, 'found:', mlog.red('NO'),
+                         '(optional)' if not required else '')
+                if required:
+                    self.is_found = False
+                    if self.required:
+                        raise DependencyException(
+                            'Could not find required LLVM Component: {}'.format(mod))
+            else:
+                self.required_modules.add(mod)
+                mlog.log('LLVM module', mod, 'found:', mlog.green('YES'))
 
     def need_threads(self):
         return True
