@@ -25,7 +25,7 @@ import argparse
 import sys, os
 import pathlib
 
-parser = argparse.ArgumentParser()
+parser = argparse.ArgumentParser(prog='meson introspect')
 parser.add_argument('--targets', action='store_true', dest='list_targets', default=False,
                     help='List top level targets.')
 parser.add_argument('--installed', action='store_true', dest='list_installed', default=False,
@@ -104,6 +104,7 @@ def list_targets(coredata, builddata, installdata):
             t['install_filename'] = determine_installed_path(target, installdata)
         else:
             t['installed'] = False
+        t['build_by_default'] = target.build_by_default
         tlist.append(t)
     print(json.dumps(tlist))
 
@@ -138,8 +139,8 @@ def add_keys(optlist, options):
         elif isinstance(opt, coredata.UserComboOption):
             optdict['choices'] = opt.choices
             typestr = 'combo'
-        elif isinstance(opt, coredata.UserStringArrayOption):
-            typestr = 'stringarray'
+        elif isinstance(opt, coredata.UserArrayOption):
+            typestr = 'array'
         else:
             raise RuntimeError("Unknown option type")
         optdict['type'] = typestr
@@ -160,9 +161,9 @@ def list_deps(coredata):
     result = []
     for d in coredata.deps.values():
         if d.found():
-            args = {'compile_args': d.get_compile_args(),
-                    'link_args': d.get_link_args()}
-            result += [d.name, args]
+            result += [{'name': d.name,
+                        'compile_args': d.get_compile_args(),
+                        'link_args': d.get_link_args()}]
     print(json.dumps(result))
 
 def list_tests(testdata):
