@@ -17,11 +17,10 @@ functionality such as gobject-introspection, gresources and gtk-doc'''
 
 from .. import build
 import os
-import sys
 import copy
 import subprocess
 from . import ModuleReturnValue
-from ..mesonlib import MesonException, OrderedSet, Popen_safe
+from ..mesonlib import MesonException, OrderedSet, Popen_safe, extract_as_list
 from ..dependencies import Dependency, PkgConfigDependency, InternalDependency
 from .. import mlog
 from .. import mesonlib
@@ -30,7 +29,7 @@ from .. import interpreter
 from . import GResourceTarget, GResourceHeaderTarget, GirTarget, TypelibTarget, VapiTarget
 from . import find_program, get_include_args
 from . import ExtensionModule
-from . import noKwargs, permittedKwargs
+from ..interpreterbase import noKwargs, permittedKwargs
 
 # gresource compilation is broken due to the way
 # the resource compiler and Ninja clash about it
@@ -98,27 +97,22 @@ class GnomeModule(ExtensionModule):
 
         cmd = ['glib-compile-resources', '@INPUT@']
 
-        source_dirs = kwargs.pop('source_dir', [])
-        if not isinstance(source_dirs, list):
-            source_dirs = [source_dirs]
+        source_dirs, dependencies = mesonlib.extract_as_list(kwargs, 'source_dir', 'dependencies', pop=True)
 
         if len(args) < 2:
             raise MesonException('Not enough arguments; the name of the resource '
                                  'and the path to the XML file are required')
 
-        dependencies = kwargs.pop('dependencies', [])
-        if not isinstance(dependencies, list):
-            dependencies = [dependencies]
         # Validate dependencies
         for (ii, dep) in enumerate(dependencies):
             if hasattr(dep, 'held_object'):
                 dependencies[ii] = dep = dep.held_object
-            if not isinstance(dep, (mesonlib.File, build.CustomTarget)):
+            if not isinstance(dep, (mesonlib.File, build.CustomTarget, build.CustomTargetIndex)):
                 m = 'Unexpected dependency type {!r} for gnome.compile_resources() ' \
                     '"dependencies" argument.\nPlease pass the return value of ' \
                     'custom_target() or configure_file()'
                 raise MesonException(m.format(dep))
-            if isinstance(dep, build.CustomTarget):
+            if isinstance(dep, (build.CustomTarget, build.CustomTargetIndex)):
                 if not mesonlib.version_compare(glib_version, gresource_dep_needed_version):
                     m = 'The "dependencies" argument of gnome.compile_resources() can not\n' \
                         'be used with the current version of glib-compile-resources due to\n' \
@@ -137,6 +131,7 @@ class GnomeModule(ExtensionModule):
         elif isinstance(ifile, str):
             ifile = os.path.join(state.subdir, ifile)
         elif isinstance(ifile, (interpreter.CustomTargetHolder,
+                                interpreter.CustomTargetIndexHolder,
                                 interpreter.GeneratedObjectsHolder)):
             m = 'Resource xml files generated at build-time cannot be used ' \
                 'with gnome.compile_resources() because we need to scan ' \
@@ -197,7 +192,7 @@ class GnomeModule(ExtensionModule):
             depfile = kwargs['output'] + '.d'
             kwargs['depfile'] = depfile
             kwargs['command'] = copy.copy(cmd) + ['--dependency-file', '@DEPFILE@']
-        target_c = GResourceTarget(name, state.subdir, kwargs)
+        target_c = GResourceTarget(name, state.subdir, state.subproject, kwargs)
 
         if gresource: # Only one target for .gresource files
             return ModuleReturnValue(target_c, [target_c])
@@ -215,7 +210,7 @@ class GnomeModule(ExtensionModule):
             h_kwargs['install'] = install_header
             h_kwargs['install_dir'] = kwargs.get('install_dir',
                                                  state.environment.coredata.get_builtin_option('includedir'))
-        target_h = GResourceHeaderTarget(args[0] + '_h', state.subdir, h_kwargs)
+        target_h = GResourceHeaderTarget(args[0] + '_h', state.subdir, state.subproject, h_kwargs)
         rv = [target_c, target_h]
         return ModuleReturnValue(rv, rv)
 
@@ -225,9 +220,10 @@ class GnomeModule(ExtensionModule):
                input_file,
                '--generate-dependencies']
 
+        # Prefer generated files over source files
+        cmd += ['--sourcedir', state.subdir] # Current build dir
         for source_dir in source_dirs:
             cmd += ['--sourcedir', os.path.join(state.subdir, source_dir)]
-        cmd += ['--sourcedir', state.subdir] # Current dir
 
         pc, stdout, stderr = Popen_safe(cmd, cwd=state.environment.get_source_dir())
         if pc.returncode != 0:
@@ -245,38 +241,32 @@ class GnomeModule(ExtensionModule):
         #
         # If there are multiple generated resource files with the same basename
         # then this code will get confused.
-
         def exists_in_srcdir(f):
             return os.path.exists(os.path.join(state.environment.get_source_dir(), f))
-        missing_dep_files = [f for f in dep_files if not exists_in_srcdir(f)]
 
         depends = []
         subdirs = []
-        for missing in missing_dep_files:
-            found = False
-            missing_basename = os.path.basename(missing)
-
+        for resfile in dep_files[:]:
+            resbasename = os.path.basename(resfile)
             for dep in dependencies:
                 if hasattr(dep, 'held_object'):
                     dep = dep.held_object
                 if isinstance(dep, mesonlib.File):
-                    if dep.fname != missing_basename:
+                    if dep.fname != resbasename:
                         continue
-                    found = True
-                    dep_files.remove(missing)
+                    dep_files.remove(resfile)
                     dep_files.append(dep)
                     subdirs.append(dep.subdir)
                     break
-                elif isinstance(dep, build.CustomTarget):
+                elif isinstance(dep, (build.CustomTarget, build.CustomTargetIndex)):
                     fname = None
                     outputs = {(o, os.path.basename(o)) for o in dep.get_outputs()}
                     for o, baseo in outputs:
-                        if baseo == missing_basename:
+                        if baseo == resbasename:
                             fname = o
                             break
                     if fname is not None:
-                        found = True
-                        dep_files.remove(missing)
+                        dep_files.remove(resfile)
                         dep_files.append(
                             mesonlib.File(
                                 is_built=True,
@@ -285,16 +275,13 @@ class GnomeModule(ExtensionModule):
                         depends.append(dep)
                         subdirs.append(dep.get_subdir())
                         break
-                else:
-                    raise RuntimeError('Unreachable code.')
-
-            if not found:
-                raise MesonException(
-                    'Resource "%s" listed in "%s" was not found. If this is a '
-                    'generated file, pass the target that generates it to '
-                    'gnome.compile_resources() using the "dependencies" '
-                    'keyword argument.' % (missing, input_file))
-
+            else:
+                if not exists_in_srcdir(resfile):
+                    raise MesonException(
+                        'Resource "%s" listed in "%s" was not found. If this is a '
+                        'generated file, pass the target that generates it to '
+                        'gnome.compile_resources() using the "dependencies" '
+                        'keyword argument.' % (resfile, input_file))
         return dep_files, depends, subdirs
 
     def _get_link_args(self, state, lib, depends=None, include_rpath=False,
@@ -328,12 +315,9 @@ class GnomeModule(ExtensionModule):
         cflags = OrderedSet()
         ldflags = OrderedSet()
         gi_includes = OrderedSet()
-        if not isinstance(deps, list):
-            deps = [deps]
+        deps = mesonlib.listify(deps, unholder=True)
 
         for dep in deps:
-            if hasattr(dep, 'held_object'):
-                dep = dep.held_object
             if isinstance(dep, InternalDependency):
                 cflags.update(get_include_args(dep.include_directories))
                 for lib in dep.libraries:
@@ -375,26 +359,32 @@ class GnomeModule(ExtensionModule):
                     # Hack to avoid passing some compiler options in
                     if lib.startswith("-W"):
                         continue
-                    if gir_has_extra_lib_arg() and use_gir_args and lib.startswith("-l"):
-                        lib = lib.replace('-l', '--extra-library=', 1)
                     ldflags.update([lib])
 
                 if isinstance(dep, PkgConfigDependency):
-                    girdir = dep.get_pkgconfig_variable("girdir")
+                    girdir = dep.get_pkgconfig_variable("girdir", {})
                     if girdir:
                         gi_includes.update([girdir])
             elif isinstance(dep, (build.StaticLibrary, build.SharedLibrary)):
                 cflags.update(get_include_args(dep.get_include_dirs()))
             else:
-                mlog.log('dependency %s not handled to build gir files' % dep)
+                mlog.log('dependency {!r} not handled to build gir files'.format(dep))
                 continue
 
+        if gir_has_extra_lib_arg() and use_gir_args:
+            fixed_ldflags = set()
+            for ldflag in ldflags:
+                if ldflag.startswith("-l"):
+                    fixed_ldflags.add(ldflag.replace('-l', '--extra-library=', 1))
+                else:
+                    fixed_ldflags.add(ldflag)
+            ldflags = fixed_ldflags
         return cflags, ldflags, gi_includes
 
     @permittedKwargs({'sources', 'nsversion', 'namespace', 'symbol_prefix', 'identifier_prefix',
                       'export_packages', 'includes', 'dependencies', 'link_with', 'include_directories',
                       'install', 'install_dir_gir', 'install_dir_typelib', 'extra_args',
-                      'packages', 'build_by_default'})
+                      'packages', 'header', 'build_by_default'})
     def generate_gir(self, state, args, kwargs):
         if len(args) != 1:
             raise MesonException('Gir takes one argument')
@@ -417,34 +407,61 @@ class GnomeModule(ExtensionModule):
             raise MesonException('gobject-introspection dependency was not found, gir cannot be generated.')
         ns = kwargs.pop('namespace')
         nsversion = kwargs.pop('nsversion')
-        libsources = kwargs.pop('sources')
+        libsources = mesonlib.extract_as_list(kwargs, 'sources', pop=True)
         girfile = '%s-%s.gir' % (ns, nsversion)
+        srcdir = os.path.join(state.environment.get_source_dir(), state.subdir)
+        builddir = os.path.join(state.environment.get_build_dir(), state.subdir)
         depends = [girtarget]
         gir_inc_dirs = []
 
-        scan_command = [giscanner, '@INPUT@']
+        scan_command = [giscanner]
         scan_command += pkgargs
         scan_command += ['--no-libtool', '--namespace=' + ns, '--nsversion=' + nsversion, '--warn-all',
                          '--output', '@OUTPUT@']
 
+        header = kwargs.pop('header', None)
+        if header:
+            if not isinstance(header, str):
+                raise MesonException('header must be a string')
+            scan_command += ['--c-include=' + header]
+
         extra_args = mesonlib.stringlistify(kwargs.pop('extra_args', []))
         scan_command += extra_args
-        scan_command += ['-I' + os.path.join(state.environment.get_source_dir(), state.subdir),
-                         '-I' + os.path.join(state.environment.get_build_dir(), state.subdir)]
+        scan_command += ['-I' + srcdir,
+                         '-I' + builddir]
         scan_command += get_include_args(girtarget.get_include_dirs())
 
+        gir_filelist_dir = state.backend.get_target_private_dir_abs(girtarget)
+        if not os.path.isdir(gir_filelist_dir):
+            os.mkdir(gir_filelist_dir)
+        gir_filelist_filename = os.path.join(gir_filelist_dir, '%s_%s_gir_filelist' % (ns, nsversion))
+
+        with open(gir_filelist_filename, 'w', encoding='utf-8') as gir_filelist:
+            for s in libsources:
+                if hasattr(s, 'held_object'):
+                    s = s.held_object
+                if isinstance(s, (build.CustomTarget, build.CustomTargetIndex)):
+                    gir_filelist.write(os.path.join(state.environment.get_build_dir(),
+                                                    state.backend.get_target_dir(s),
+                                                    s.get_outputs()[0]) + '\n')
+                elif isinstance(s, mesonlib.File):
+                    gir_filelist.write(s.rel_to_builddir(state.build_to_src) + '\n')
+                elif isinstance(s, build.GeneratedList):
+                    for gen_src in s.get_outputs():
+                        gir_filelist.write(os.path.join(srcdir, gen_src) + '\n')
+                else:
+                    gir_filelist.write(os.path.join(srcdir, s) + '\n')
+        scan_command += ['--filelist=' + gir_filelist_filename]
+
         if 'link_with' in kwargs:
-            link_with = kwargs.pop('link_with')
-            if not isinstance(link_with, list):
-                link_with = [link_with]
+            link_with = mesonlib.extract_as_list(kwargs, 'link_with', pop = True)
+
             for link in link_with:
                 scan_command += self._get_link_args(state, link.held_object, depends,
                                                     use_gir_args=True)
 
         if 'includes' in kwargs:
-            includes = kwargs.pop('includes')
-            if not isinstance(includes, list):
-                includes = [includes]
+            includes = mesonlib.extract_as_list(kwargs, 'includes', pop = True)
             for inc in includes:
                 if hasattr(inc, 'held_object'):
                     inc = inc.held_object
@@ -456,7 +473,7 @@ class GnomeModule(ExtensionModule):
                                      inc.get_subdir()),
                     ]
                     scan_command += [
-                        "--include=%s" % (inc.get_basename()[:-4], ),
+                        "--include-uninstalled=%s" % (os.path.join(inc.get_subdir(), inc.get_basename()), )
                     ]
                     depends += [inc]
                 else:
@@ -485,17 +502,15 @@ class GnomeModule(ExtensionModule):
                 # FIXME: Linking directly to libasan is not recommended but g-ir-scanner
                 # does not understand -f LDFLAGS. https://bugzilla.gnome.org/show_bug.cgi?id=783892
                 # ldflags += compilers.sanitizer_link_args(sanitize)
-        if kwargs.get('symbol_prefix'):
-            sym_prefix = kwargs.pop('symbol_prefix')
-            if not isinstance(sym_prefix, str):
-                raise MesonException('Gir symbol prefix must be str')
-            scan_command += ['--symbol-prefix=%s' % sym_prefix]
-        if kwargs.get('identifier_prefix'):
+        if 'symbol_prefix' in kwargs:
+            sym_prefixes = mesonlib.stringlistify(kwargs.pop('symbol_prefix', []))
+            scan_command += ['--symbol-prefix=%s' % sym_prefix for sym_prefix in sym_prefixes]
+        if 'identifier_prefix' in kwargs:
             identifier_prefix = kwargs.pop('identifier_prefix')
             if not isinstance(identifier_prefix, str):
                 raise MesonException('Gir identifier prefix must be str')
             scan_command += ['--identifier-prefix=%s' % identifier_prefix]
-        if kwargs.get('export_packages'):
+        if 'export_packages' in kwargs:
             pkgs = kwargs.pop('export_packages')
             if isinstance(pkgs, str):
                 scan_command += ['--pkg-export=%s' % pkgs]
@@ -504,11 +519,8 @@ class GnomeModule(ExtensionModule):
             else:
                 raise MesonException('Gir export packages must be str or list')
 
-        deps = kwargs.pop('dependencies', [])
-        if not isinstance(deps, list):
-            deps = [deps]
         deps = (girtarget.get_all_link_deps() + girtarget.get_external_deps() +
-                deps)
+                extract_as_list(kwargs, 'dependencies', pop=True, unholder=True))
         # Need to recursively add deps on GirTarget sources from our
         # dependencies and also find the include directories needed for the
         # typelib generation custom target below.
@@ -541,7 +553,7 @@ class GnomeModule(ExtensionModule):
                         if subdir not in typelib_includes:
                             typelib_includes.append(subdir)
             elif isinstance(dep, PkgConfigDependency):
-                girdir = dep.get_pkgconfig_variable("girdir")
+                girdir = dep.get_pkgconfig_variable("girdir", {})
                 if girdir and girdir not in typelib_includes:
                     typelib_includes.append(girdir)
         # ldflags will be misinterpreted by gir scanner (showing
@@ -563,9 +575,7 @@ class GnomeModule(ExtensionModule):
         for i in gi_includes:
             scan_command += ['--add-include-path=%s' % i]
 
-        inc_dirs = kwargs.pop('include_directories', [])
-        if not isinstance(inc_dirs, list):
-            inc_dirs = [inc_dirs]
+        inc_dirs = mesonlib.extract_as_list(kwargs, 'include_directories', pop = True)
         for incd in inc_dirs:
             if not isinstance(incd.held_object, (str, build.IncludeDirs)):
                 raise MesonException(
@@ -586,16 +596,15 @@ class GnomeModule(ExtensionModule):
                 scan_command.append('-L' + d)
             scan_command += ['--library', libname]
         scankwargs = {'output': girfile,
-                      'input': libsources,
                       'command': scan_command,
                       'depends': depends}
-        if kwargs.get('install'):
+        if 'install' in kwargs:
             scankwargs['install'] = kwargs['install']
             scankwargs['install_dir'] = kwargs.get('install_dir_gir',
                                                    os.path.join(state.environment.get_datadir(), 'gir-1.0'))
         if 'build_by_default' in kwargs:
             scankwargs['build_by_default'] = kwargs['build_by_default']
-        scan_target = GirTarget(girfile, state.subdir, scankwargs)
+        scan_target = GirTarget(girfile, state.subdir, state.subproject, scankwargs)
 
         typelib_output = '%s-%s.typelib' % (ns, nsversion)
         typelib_cmd = [gicompiler, scan_target, '--output', '@OUTPUT@']
@@ -607,13 +616,13 @@ class GnomeModule(ExtensionModule):
             'output': typelib_output,
             'command': typelib_cmd,
         }
-        if kwargs.get('install'):
+        if 'install' in kwargs:
             typelib_kwargs['install'] = kwargs['install']
             typelib_kwargs['install_dir'] = kwargs.get('install_dir_typelib',
                                                        os.path.join(state.environment.get_libdir(), 'girepository-1.0'))
         if 'build_by_default' in kwargs:
             typelib_kwargs['build_by_default'] = kwargs['build_by_default']
-        typelib_target = TypelibTarget(typelib_output, state.subdir, typelib_kwargs)
+        typelib_target = TypelibTarget(typelib_output, state.subdir, state.subproject, typelib_kwargs)
         rv = [scan_target, typelib_target]
         return ModuleReturnValue(rv, rv)
 
@@ -632,8 +641,8 @@ class GnomeModule(ExtensionModule):
         if state.subdir == '':
             targetname = 'gsettings-compile'
         else:
-            targetname = 'gsettings-compile-' + state.subdir
-        target_g = build.CustomTarget(targetname, state.subdir, kwargs)
+            targetname = 'gsettings-compile-' + state.subdir.replace('/', '_')
+        target_g = build.CustomTarget(targetname, state.subdir, state.subproject, kwargs)
         return ModuleReturnValue(target_g, [target_g])
 
     @permittedKwargs({'sources', 'media', 'symlink_media', 'languages'})
@@ -651,6 +660,11 @@ class GnomeModule(ExtensionModule):
         source_str = '@@'.join(sources)
 
         langs = mesonlib.stringlistify(kwargs.pop('languages', []))
+        if langs:
+            mlog.log(mlog.red('DEPRECATION:'), '''The "languages" argument of gnome.yelp() is deprecated.
+Use a LINGUAS file in the sources directory instead.
+This will become a hard error in the future.''')
+
         media = mesonlib.stringlistify(kwargs.pop('media', []))
         symlinks = kwargs.pop('symlink_media', True)
 
@@ -660,7 +674,7 @@ class GnomeModule(ExtensionModule):
         if kwargs:
             raise MesonException('Unknown arguments passed: {}'.format(', '.join(kwargs.keys())))
 
-        script = [sys.executable, state.environment.get_build_command()]
+        script = state.environment.get_build_command()
         args = ['--internal',
                 'yelphelper',
                 'install',
@@ -676,20 +690,24 @@ class GnomeModule(ExtensionModule):
             args.append('--langs=' + '@@'.join(langs))
         inscript = build.RunScript(script, args)
 
-        potargs = [state.environment.get_build_command(), '--internal', 'yelphelper', 'pot',
-                   '--subdir=' + state.subdir,
-                   '--id=' + project_id,
-                   '--sources=' + source_str]
-        pottarget = build.RunTarget('help-' + project_id + '-pot', sys.executable,
-                                    potargs, [], state.subdir)
+        potargs = state.environment.get_build_command() + [
+            '--internal', 'yelphelper', 'pot',
+            '--subdir=' + state.subdir,
+            '--id=' + project_id,
+            '--sources=' + source_str,
+        ]
+        pottarget = build.RunTarget('help-' + project_id + '-pot', potargs[0],
+                                    potargs[1:], [], state.subdir, state.subproject)
 
-        poargs = [state.environment.get_build_command(), '--internal', 'yelphelper', 'update-po',
-                  '--subdir=' + state.subdir,
-                  '--id=' + project_id,
-                  '--sources=' + source_str,
-                  '--langs=' + '@@'.join(langs)]
-        potarget = build.RunTarget('help-' + project_id + '-update-po', sys.executable,
-                                   poargs, [], state.subdir)
+        poargs = state.environment.get_build_command() + [
+            '--internal', 'yelphelper', 'update-po',
+            '--subdir=' + state.subdir,
+            '--id=' + project_id,
+            '--sources=' + source_str,
+            '--langs=' + '@@'.join(langs),
+        ]
+        potarget = build.RunTarget('help-' + project_id + '-update-po', poargs[0],
+                                   poargs[1:], [], state.subdir, state.subproject)
 
         rv = [inscript, pottarget, potarget]
         return ModuleReturnValue(None, rv)
@@ -697,7 +715,8 @@ class GnomeModule(ExtensionModule):
     @permittedKwargs({'main_xml', 'main_sgml', 'src_dir', 'dependencies', 'install',
                       'install_dir', 'scan_args', 'scanobjs_args', 'gobject_typesfile',
                       'fixxref_args', 'html_args', 'html_assets', 'content_files',
-                      'mkdb_args', 'ignore_headers'})
+                      'mkdb_args', 'ignore_headers', 'include_directories',
+                      'namespace', 'mode', 'expand_content_files'})
     def gtkdoc(self, state, args, kwargs):
         if len(args) != 1:
             raise MesonException('Gtkdoc must have one positional argument.')
@@ -717,7 +736,7 @@ class GnomeModule(ExtensionModule):
                 raise MesonException('You can only specify main_xml or main_sgml, not both.')
             main_file = main_xml
         targetname = modulename + '-doc'
-        command = [sys.executable, state.environment.get_build_command()]
+        command = state.environment.get_build_command()
 
         namespace = kwargs.get('namespace', '')
         mode = kwargs.get('mode', 'auto')
@@ -725,9 +744,7 @@ class GnomeModule(ExtensionModule):
         if mode not in VALID_MODES:
             raise MesonException('gtkdoc: Mode {} is not a valid mode: {}'.format(mode, VALID_MODES))
 
-        src_dirs = kwargs['src_dir']
-        if not isinstance(src_dirs, list):
-            src_dirs = [src_dirs]
+        src_dirs = mesonlib.extract_as_list(kwargs, 'src_dir')
         header_dirs = []
         for src_dir in src_dirs:
             if hasattr(src_dir, 'held_object'):
@@ -764,17 +781,16 @@ class GnomeModule(ExtensionModule):
         args += self._unpack_args('--ignore-headers=', 'ignore_headers', kwargs)
         args += self._unpack_args('--installdir=', 'install_dir', kwargs, state)
         args += self._get_build_args(kwargs, state)
-        res = [build.RunTarget(targetname, command[0], command[1:] + args, [], state.subdir)]
+        res = [build.RunTarget(targetname, command[0], command[1:] + args, [], state.subdir, state.subproject)]
         if kwargs.get('install', True):
             res.append(build.RunScript(command, args))
         return ModuleReturnValue(None, res)
 
     def _get_build_args(self, kwargs, state):
         args = []
-        cflags, ldflags, gi_includes = self._get_dependencies_flags(kwargs.get('dependencies', []), state, include_rpath=True)
-        inc_dirs = kwargs.get('include_directories', [])
-        if not isinstance(inc_dirs, list):
-            inc_dirs = [inc_dirs]
+        deps = extract_as_list(kwargs, 'dependencies', unholder=True)
+        cflags, ldflags, gi_includes = self._get_dependencies_flags(deps, state, include_rpath=True)
+        inc_dirs = mesonlib.extract_as_list(kwargs, 'include_directories')
         for incd in inc_dirs:
             if not isinstance(incd.held_object, (str, build.IncludeDirs)):
                 raise MesonException(
@@ -805,13 +821,11 @@ class GnomeModule(ExtensionModule):
         if kwarg_name not in kwargs:
             return []
 
-        new_args = kwargs[kwarg_name]
-        if not isinstance(new_args, list):
-            new_args = [new_args]
+        new_args = mesonlib.extract_as_list(kwargs, kwarg_name)
         args = []
         for i in new_args:
             if expend_file_state and isinstance(i, mesonlib.File):
-                i = os.path.join(expend_file_state.environment.get_build_dir(), i.subdir, i.fname)
+                i = i.absolute_path(expend_file_state.environment.get_source_dir(), expend_file_state.environment.get_build_dir())
             elif not isinstance(i, str):
                 raise MesonException(kwarg_name + ' values must be strings.')
             args.append(i)
@@ -821,7 +835,8 @@ class GnomeModule(ExtensionModule):
 
         return []
 
-    @permittedKwargs({'interface_prefix', 'namespace', 'object_manager', 'build_by_default'})
+    @permittedKwargs({'interface_prefix', 'namespace', 'object_manager', 'build_by_default',
+                      'annotations', 'docbook'})
     def gdbus_codegen(self, state, args, kwargs):
         if len(args) != 2:
             raise MesonException('Gdbus_codegen takes two arguments, name and xml file.')
@@ -835,6 +850,20 @@ class GnomeModule(ExtensionModule):
             cmd += ['--c-namespace', kwargs.pop('namespace')]
         if kwargs.get('object_manager', False):
             cmd += ['--c-generate-object-manager']
+        if 'docbook' in kwargs:
+            cmd += ['--generate-docbook', kwargs.pop('docbook')]
+
+        # Annotations are a bit ugly in that they are a list of lists of strings...
+        annotations = kwargs.pop('annotations', [])
+        if not isinstance(annotations, list):
+            raise MesonException('annotations takes a list')
+        if annotations and isinstance(annotations, list) and not isinstance(annotations[0], list):
+            annotations = [annotations]
+
+        for annotation in annotations:
+            if len(annotation) != 3 or not all(isinstance(i, str) for i in annotation):
+                raise MesonException('Annotations must be made up of 3 strings for ELEMENT, KEY, and VALUE')
+            cmd += ['--annotate'] + annotation
 
         # https://git.gnome.org/browse/glib/commit/?id=ee09bb704fe9ccb24d92dd86696a0e6bb8f0dc1a
         if mesonlib.version_compare(self._get_native_glib_version(state), '>= 2.51.3'):
@@ -849,7 +878,7 @@ class GnomeModule(ExtensionModule):
                          }
         if 'build_by_default' in kwargs:
             custom_kwargs['build_by_default'] = kwargs['build_by_default']
-        ct = build.CustomTarget(target_name, state.subdir, custom_kwargs)
+        ct = build.CustomTarget(target_name, state.subdir, state.subproject, custom_kwargs)
         return ModuleReturnValue(ct, [ct])
 
     @permittedKwargs({'sources', 'c_template', 'h_template', 'install_header', 'install_dir',
@@ -1065,7 +1094,7 @@ G_END_DECLS'''
             'command': cmd
         }
         custom_kwargs.update(kwargs)
-        return build.CustomTarget(output, state.subdir, custom_kwargs,
+        return build.CustomTarget(output, state.subdir, state.subproject, custom_kwargs,
                                   # https://github.com/mesonbuild/meson/issues/973
                                   absolute_paths=True)
 
@@ -1135,7 +1164,7 @@ G_END_DECLS'''
             # Silence any warnings about missing prototypes
             custom_kwargs['command'] += ['--include-header', header_file]
         custom_kwargs['output'] = output + '.c'
-        body = build.CustomTarget(output + '_c', state.subdir, custom_kwargs)
+        body = build.CustomTarget(output + '_c', state.subdir, state.subproject, custom_kwargs)
 
         custom_kwargs['install'] = install_header
         if install_dir is not None:
@@ -1144,19 +1173,15 @@ G_END_DECLS'''
             cmd += ['--pragma-once']
         custom_kwargs['command'] = cmd + ['--header', '@INPUT@']
         custom_kwargs['output'] = header_file
-        header = build.CustomTarget(output + '_h', state.subdir, custom_kwargs)
+        header = build.CustomTarget(output + '_h', state.subdir, state.subproject, custom_kwargs)
 
         rv = [body, header]
         return ModuleReturnValue(rv, rv)
 
     @staticmethod
     def _vapi_args_to_command(prefix, variable, kwargs, accept_vapi=False):
-        arg_list = kwargs.get(variable)
-        if not arg_list:
-            return []
+        arg_list = mesonlib.extract_as_list(kwargs, variable)
         ret = []
-        if not isinstance(arg_list, list):
-            arg_list = [arg_list]
         for arg in arg_list:
             if not isinstance(arg, str):
                 types = 'strings' + ' or InternalDependencys' if accept_vapi else ''
@@ -1175,9 +1200,7 @@ G_END_DECLS'''
         arg_list = kwargs.get('packages')
         if not arg_list:
             return [], [], [], []
-        if not isinstance(arg_list, list):
-            arg_list = [arg_list]
-
+        arg_list = mesonlib.listify(arg_list)
         vapi_depends = []
         vapi_packages = []
         vapi_includes = []
@@ -1251,12 +1274,10 @@ G_END_DECLS'''
         cmd += pkg_cmd
         cmd += ['--metadatadir=' + source_dir]
 
-        inputs = kwargs.get('sources')
-        if not inputs:
+        if 'sources' not in kwargs:
             raise MesonException('sources are required to generate the vapi file')
 
-        if not isinstance(inputs, list):
-            inputs = [inputs]
+        inputs = mesonlib.extract_as_list(kwargs, 'sources')
 
         link_with = []
         for i in inputs:
@@ -1288,7 +1309,7 @@ G_END_DECLS'''
             # We shouldn't need this locally but we install it
             deps_target = self._generate_deps(state, library, vapi_packages, install_dir)
             created_values.append(deps_target)
-        vapi_target = VapiTarget(vapi_output, state.subdir, custom_kwargs)
+        vapi_target = VapiTarget(vapi_output, state.subdir, state.subproject, custom_kwargs)
 
         # So to try our best to get this to just work we need:
         # - link with with the correct library
